# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import pathlib
import shutil
import socket
import subprocess
import sys
import tempfile
import time

import pytest


@pytest.fixture(scope="module")
def temp_dir():
    tmp = pathlib.Path(tempfile.mkdtemp())
    yield tmp
    shutil.rmtree(tmp)


@pytest.fixture(scope="module")
def dlrepo_server(request, temp_dir):  # pylint: disable=redefined-outer-name
    data_dir = pathlib.Path(request.fspath).parent / request.module.__name__
    for d in ("branches", "products", "users"):
        folder = data_dir / d
        if folder.is_dir():
            shutil.copytree(folder, temp_dir / d)
    portnum = get_free_tcp_port()
    env = {
        "DLREPO_ROOT_DIR": str(temp_dir),
        "DLREPO_LISTEN_ADDRESS": "127.0.0.1",
        "DLREPO_LISTEN_PORT": str(portnum),
        "DLREPO_TEMPLATES_DIR": str(data_dir / "templates"),
        "DLREPO_STATIC_DIR": str(data_dir / "static"),
        "DLREPO_LOG_OUTPUT": "console",
        "DLREPO_LOG_LEVEL": "DEBUG",
    }
    acls = data_dir / "acls"
    if acls.is_dir():
        env["DLREPO_ACLS_DIR"] = str(acls)
    auth = data_dir / "auth"
    if auth.is_file():
        env["DLREPO_AUTH_FILE"] = str(auth)
    else:
        headers = data_dir / "headers"
        if headers.is_dir():
            for h in headers.iterdir():
                env[h.name] = h.read_text("utf-8").strip()
        else:
            env["DLREPO_AUTH_DISABLED"] = "1"
    settings = data_dir / "settings"
    if settings.is_dir():
        for s in settings.iterdir():
            env[s.name] = s.read_text("utf-8").strip()
    post_process = data_dir / "post-process.sh"
    if post_process.is_file():
        env["DLREPO_POST_PROCESS_CMD"] = str(post_process)
    with subprocess.Popen([sys.executable, "-m", "dlrepo"], env=env) as proc:
        while proc.poll() is None:
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.connect(("127.0.0.1", portnum))
                break
            except socket.error:
                time.sleep(0.1)
        assert proc.poll() is None
        try:
            yield f"http://127.0.0.1:{portnum}", temp_dir
        finally:
            proc.terminate()


def get_free_tcp_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(("127.0.0.1", 0))
        _, portnum = sock.getsockname()
        return portnum



# Copyright (c) 2022 Julien Floret
# Copyright (c) 2022 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import io
import tarfile

import aiohttp
import pytest


pytestmark = pytest.mark.asyncio


async def test_archive(dlrepo_server):
    url, _ = dlrepo_server
    async with aiohttp.ClientSession(url) as sess:
        data2 = b"content2"
        digest2 = hashlib.sha256(data2).hexdigest()
        resp = await sess.put(
            "/branches/main/tag/job/format/file2.txt",
            data=data2,
            headers={"Digest": f"sha256:{digest2}"},
        )
        assert resp.status == 201

        data = b"content"
        digest = hashlib.sha256(data).hexdigest()
        resp = await sess.put(
            "/branches/main/tag/job/format/file.txt",
            data=data,
            headers={"Digest": f"sha256:{digest}"},
        )
        assert resp.status == 201

        resp = await sess.get("/branches/main/tag/job/format.tar")
        assert resp.status == 200
        assert "Content-Type" in resp.headers
        assert resp.headers["Content-Type"] == "application/x-tar"
        tar = await resp.read()
        with tarfile.open(fileobj=io.BytesIO(tar)) as t:
            assert t.getnames() == [
                "job-tag-format/file.txt",
                "job-tag-format/file2.txt",
                "job-tag-format/SHA256SUMS.txt",
            ]
            with t.extractfile("job-tag-format/file.txt") as f:
                assert f.read() == data
            with t.extractfile("job-tag-format/file2.txt") as f:
                assert f.read() == data2
            with t.extractfile("job-tag-format/SHA256SUMS.txt") as f:
                txt = f.read().decode("utf-8")
                assert txt == f"{digest}  file.txt\n{digest2}  file2.txt\n"



# Copyright (c) 2022 Julien Floret
# Copyright (c) 2022 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
import pathlib
import shutil
import socket
import subprocess
import sys
import tempfile
import time

import aiohttp
import pytest


reference = pathlib.Path(__file__).parent / "publish_reference"
pytestmark = pytest.mark.asyncio


@pytest.fixture(scope="module", params=["1", "10"])
def dlrepo_servers(request, dlrepo_server):
    tmp = pathlib.Path(tempfile.mkdtemp())
    shutil.copytree(reference / "branches", tmp / "branches")
    shutil.copytree(reference / "products", tmp / "products", symlinks=True)
    auth = tmp / "publish.auth"
    auth.write_text("foo:bar")  # use garbage credentials, auth is disabled anyway
    public_url, _ = dlrepo_server

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(("127.0.0.1", 0))
        _, portnum = sock.getsockname()

    env = {
        "DLREPO_ROOT_DIR": str(tmp),
        "DLREPO_LISTEN_ADDRESS": "127.0.0.1",
        "DLREPO_LISTEN_PORT": str(portnum),
        "DLREPO_LOG_OUTPUT": "console",
        "DLREPO_LOG_LEVEL": "DEBUG",
        "DLREPO_AUTH_DISABLED": "1",
        "DLREPO_PUBLISH_URL": public_url,
        "DLREPO_PUBLISH_AUTH": auth,
        "DLREPO_PUBLISH_MAX_REQUESTS": request.param,
    }
    with subprocess.Popen([sys.executable, "-m", "dlrepo"], env=env) as proc:
        while proc.poll() is None:
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.connect(("127.0.0.1", portnum))
                break
            except socket.error:
                time.sleep(0.1)
        assert proc.poll() is None
        try:
            yield f"http://127.0.0.1:{portnum}", public_url
        finally:
            proc.terminate()

    shutil.rmtree(tmp)


async def test_publish(dlrepo_servers):  # pylint: disable=redefined-outer-name
    url, public_url = dlrepo_servers
    async with aiohttp.ClientSession(url) as sess:
        resp = await sess.put("/branches/branch/tag/", json={"tag": {"released": True}})
        assert resp.status == 200
        await asyncio.sleep(1)
        resp = await sess.get("/branches/branch/tag/")
        assert resp.status == 200
        data = await resp.json()
        assert data["tag"]["publish_status"] == f"published to {public_url}"

        async with aiohttp.ClientSession(public_url) as pub_sess:
            for job in "job1", "job2":
                for fmt in "fmt1", "fmt2":
                    sha_url = f"/branches/branch/tag/{job}/{fmt}.sha256"
                    resp = await sess.get(sha_url)
                    assert resp.status == 200
                    data = await resp.text()
                    resp = await pub_sess.get(sha_url)
                    assert resp.status == 200
                    data_pub = await resp.text()
                    assert data_pub == data
            sha_url = "/branches/branch/tag/internal_job/fmt1.sha256"
            resp = await sess.get(sha_url)
            assert resp.status == 200
            resp = await pub_sess.get(sha_url)
            assert resp.status == 404



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import os
from pathlib import Path
import re
from typing import Iterator, Tuple, Union


# --------------------------------------------------------------------------------------
class SubDir:
    def __init__(self, parent: Union["AbstractRepository", "SubDir"], name: str):
        self.parent = parent
        self.name = name.lower()
        self._path = self._resolve_path()

    def root(self) -> "AbstractRepository":
        r = self
        while r.parent is not None:
            r = r.parent
        return r

    def url_bit(self) -> str:
        return self.name

    def url(self) -> str:
        bits = [self.url_bit()]
        r = self
        while r.parent is not None:
            r = r.parent
            if r.url_bit() is not None:
                bits.insert(0, r.url_bit())
        return f"/{'/'.join(bits)}/"

    @classmethod
    def all(cls, parent) -> Iterator["SubDir"]:
        try:
            dirs = list(cls.parent_path(parent).iterdir())
            dirs.sort(key=lambda d: d.name)
            for d in dirs:
                if d.name.startswith(".") or not d.is_dir():
                    continue
                yield cls(parent, d.name)
        except FileNotFoundError:
            pass

    @classmethod
    def parent_path(cls, parent: "SubDir") -> Path:
        return parent.path()

    def _resolve_path(self) -> Path:
        path = self.parent_path(self.parent) / self.name
        if self.name.startswith(".") or "/" in self.name:
            raise FileNotFoundError(path)
        return path

    def exists(self) -> bool:
        return self.path().is_dir()

    def create(self):
        if self.parent is not None:
            self.parent.create()
        self.path().mkdir(mode=0o755, parents=True, exist_ok=True)

    def path(self) -> Path:
        return self._path


# --------------------------------------------------------------------------------------
HASH_ALGOS = "|".join(hashlib.algorithms_guaranteed)
ALGO_RE = re.compile(
    rf"""
    ^
    (?P<algo>{HASH_ALGOS})
    :
    (?P<digest>[A-Fa-f0-9]+)
    $
    """,
    re.VERBOSE,
)


def parse_digest(digest: str) -> Tuple[str, str]:
    match = ALGO_RE.match(digest)
    if not match:
        raise ValueError(f"invalid digest: {digest}")
    return match.groups()


# --------------------------------------------------------------------------------------
CHUNK_SIZE = int(os.getenv("DLREPO_CHUNK_SIZE", str(256 * 1024)))


def file_digest(algo: str, path: Path) -> str:
    h = hashlib.new(algo)
    buf = bytearray(CHUNK_SIZE)
    view = memoryview(buf)
    with path.open("rb") as f:
        while True:
            n = f.readinto(buf)
            if not n:
                break
            h.update(view[:n])
    return h.hexdigest()


# --------------------------------------------------------------------------------------
def human_readable(value):
    if value == 0:
        return "0"
    units = ("K", "M", "G", "T")
    i = 0
    unit = ""
    while value >= 1000 and i < len(units):
        unit = units[i]
        value /= 1000
        i += 1
    if value < 100:
        return f"{value:.1f}{unit}"
    return f"{value:.0f}{unit}"



# Copyright (c) 2022 Julien Floret
# Copyright (c) 2022 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import hashlib

import aiohttp
import pytest


pytestmark = pytest.mark.asyncio


async def test_ok(dlrepo_server):
    url, _ = dlrepo_server
    async with aiohttp.ClientSession(url) as sess:
        resp = await sess.get("/branches/branch/tag/job/fmt/file.txt")
        assert resp.status == 200
        data = await resp.read()
        assert data == b"foo\n"
        resp = await sess.head("/branches/branch/tag/job/fmt/finalized")
        assert resp.status == 404
        resp = await sess.patch("/branches/branch/tag/job/fmt/")
        assert resp.status == 200
        resp = await sess.get("/branches/branch/tag/job/fmt/file.txt")
        assert resp.status == 200
        data = await resp.read()
        assert data == b"foo\nfinalized\n"
        resp = await sess.get("/branches/branch/tag/job/fmt/finalized")
        assert resp.status == 200


async def test_ignored(dlrepo_server):
    url, _ = dlrepo_server
    async with aiohttp.ClientSession(url) as sess:
        resp = await sess.get("/branches/branch/tag/job/fmt-ignored/foo.txt")
        assert "Digest" not in resp.headers
        assert resp.status == 200
        data = await resp.read()
        assert data == b"foobar\n"
        resp = await sess.patch("/branches/branch/tag/job/fmt-ignored/")
        assert resp.status == 200
        resp = await sess.get("/branches/branch/tag/job/fmt-ignored/foo.txt")
        assert resp.status == 200
        data = await resp.read()
        assert data == b"foobar\n"
        digest = hashlib.sha256(data).hexdigest()
        assert "Digest" in resp.headers
        assert resp.headers["Digest"] == f"sha256:{digest}"


async def test_error(dlrepo_server):
    url, _ = dlrepo_server
    async with aiohttp.ClientSession(url) as sess:
        resp = await sess.get("/branches/branch/tag/job/fmt-error/file3.txt")
        assert resp.status == 200
        data = await resp.read()
        assert data == b"foobar\n"
        resp = await sess.patch("/branches/branch/tag/job/fmt-error/")
        assert resp.status == 500
        resp = await sess.get("/branches/branch/tag/job/fmt-error/file3.txt")
        assert resp.status == 200
        data = await resp.read()
        assert data == b"foobar\n"



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import hashlib

import aiohttp
import pytest


pytestmark = pytest.mark.asyncio


async def test_noauth(dlrepo_server):
    url, _ = dlrepo_server
    async with aiohttp.ClientSession(url) as sess:
        for u in ("/", "/not-found", "/cli"):
            resp = await sess.get(u)
            assert resp.status == 401
        resp = await sess.get("/static/logo.svg")
        assert resp.status == 200


async def test_invalid_auth(dlrepo_server):
    url, _ = dlrepo_server
    auth = aiohttp.BasicAuth("foo", "invalidpassword")
    async with aiohttp.ClientSession(url, auth=auth) as sess:
        for u in ("/", "/not-found", "/cli"):
            resp = await sess.get(u)
            assert resp.status == 401
        resp = await sess.get("/static/logo.svg")
        assert resp.status == 200


async def test_valid_auth(dlrepo_server):
    url, _ = dlrepo_server
    auth = aiohttp.BasicAuth("foo", "baz")
    async with aiohttp.ClientSession(url) as sess:
        resp = await sess.get("/", auth=auth)
        assert resp.status == 200
        resp = await sess.get("/not-found", auth=auth)
        assert resp.status == 404
        resp = await sess.get("/cli", auth=auth)
        assert resp.status == 200
        text = await resp.text()
        with open("dlrepo-cli", "r", encoding="utf-8") as f:
            assert text == f.read()
        resp = await sess.get("/static/logo.svg", auth=auth)
        assert resp.status == 200


async def test_write_denied(dlrepo_server):
    url, _ = dlrepo_server
    auth = aiohttp.BasicAuth("foo", "baz")
    async with aiohttp.ClientSession(url, auth=auth) as sess:
        resp = await sess.put("/branches/main/tag/job/format/file.txt")
        assert resp.status == 401


async def test_write_allowed(dlrepo_server):
    url, _ = dlrepo_server
    auth = aiohttp.BasicAuth("bar", "bar")
    async with aiohttp.ClientSession(url, auth=auth) as sess:
        data = b"content"
        digest = hashlib.sha256(data).hexdigest()
        resp = await sess.put(
            "/branches/main/tag/job/format/file.txt",
            data=data,
            headers={"Digest": f"sha256:{digest}"},
        )
        assert resp.status == 201
        resp = await sess.put(
            "/branches/main/tag/job/format/file2.txt",
            data=data + b"corrupt",
            headers={"Digest": f"sha256:{digest}"},
        )
        assert resp.status == 422


async def test_read_allowed(dlrepo_server):
    url, _ = dlrepo_server
    auth = aiohttp.BasicAuth("foo", "baz")
    async with aiohttp.ClientSession(url, auth=auth) as sess:
        resp = await sess.get("/branches/main/tag/job/format/file.txt")
        assert resp.status == 200
        data = await resp.read()
        assert data == b"content"


async def test_read_denied(dlrepo_server):
    url, _ = dlrepo_server
    async with aiohttp.ClientSession(url) as sess:
        resp = await sess.get("/branches/main/tag/job/format/file.txt")
        assert resp.status == 401
    auth = aiohttp.BasicAuth("coin", "coin")
    async with aiohttp.ClientSession(url, auth=auth) as sess:
        resp = await sess.get("/branches/main/tag/job/format/file.txt")
        assert resp.status == 401


async def test_acl_regexp(dlrepo_server):
    url, _ = dlrepo_server
    auth = aiohttp.BasicAuth("coin", "coin")
    async with aiohttp.ClientSession(url, auth=auth) as sess:
        data = b"content"
        digest = hashlib.sha256(data).hexdigest()
        resp = await sess.put(
            "/branches/main/tag2/job/format/file2.txt",
            data=data,
            headers={"Digest": f"sha256:{digest}"},
        )
        assert resp.status == 201
        resp = await sess.get("/branches/main/tag2/job/format/file2.txt")
        assert resp.status == 200
        data = await resp.read()
        assert data == b"content"



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import logging
from pathlib import Path
from typing import Callable, Iterator

from .fmt import ArtifactFormat
from .util import SubDir


LOG = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------
class Product(SubDir):
    """
    TODO
    """

    ROOT_DIR = "products"

    @classmethod
    def parent_path(cls, parent: "ArtifactRepository") -> Path:
        return parent.path() / cls.ROOT_DIR

    def url_bit(self) -> str:
        return f"products/{self.name}"

    def get_variants(self) -> Iterator["Variant"]:
        yield from Variant.all(self)

    def get_variant(self, name: str) -> "Variant":
        return Variant(self, name)


# --------------------------------------------------------------------------------------
class Variant(SubDir):
    """
    TODO
    """

    def get_branches(self) -> Iterator["ProductBranch"]:
        yield from ProductBranch.all(self)

    def get_branch(self, name: str) -> "ProductBranch":
        return ProductBranch(self, name)


# --------------------------------------------------------------------------------------
class ProductBranch(SubDir):
    """
    TODO
    """

    def get_versions(
        self, access_cb: Callable[[str], bool] = None
    ) -> Iterator["Version"]:
        for v in Version.all(self):
            if access_cb is not None and not access_cb(v.url()):
                continue
            yield v

    def get_version(
        self, name: str, access_cb: Callable[[str], bool] = None
    ) -> "Version":
        if name in ("latest", "stable", "oldstable"):
            versions = list(self.get_versions(access_cb))
            versions.sort(key=Version.creation_date, reverse=True)
            first = True
            for v in versions:
                if name == "latest":
                    return v
                if v.is_stable():
                    if name == "stable" or not first:
                        return v
                    first = False
            raise FileNotFoundError(name)
        return Version(self, name)


# --------------------------------------------------------------------------------------
class Version(SubDir):
    """
    TODO
    """

    def create(self):
        super().create()
        stamp = self.path() / ".stamp"
        if not stamp.exists():
            stamp.touch()

    def archive_name(self) -> str:
        variant = self.parent.parent
        product = variant.parent
        return f"{product.name}-{variant.name}-v{self.name}"

    @classmethod
    def creation_date(cls, v):
        stamp = v.path() / ".stamp"
        if stamp.is_file():
            # prefer mtime over ctime
            # on UNIX, ctime is "the time of most recent metadata change" whereas
            # mtime is "most recent content modification"
            return stamp.stat().st_mtime
        return 0

    @property
    def timestamp(self) -> int:
        return Version.creation_date(self)

    def is_released(self) -> bool:
        for fmt in self.get_formats():
            released_path = fmt.path().resolve().parent / ".released"
            if released_path.is_file():
                return True
        return False

    def is_locked(self) -> bool:
        for fmt in self.get_formats():
            locked_path = fmt.path().resolve().parent.parent / ".locked"
            if locked_path.is_file():
                return True
        return False

    def is_stable(self) -> bool:
        for fmt in self.get_formats():
            stable_path = fmt.path().resolve().parent.parent / ".stable"
            if stable_path.is_file():
                return True
        return False

    def get_formats(self) -> Iterator[ArtifactFormat]:
        yield from ArtifactFormat.all(self)

    def get_format(self, name: str) -> ArtifactFormat:
        return ArtifactFormat(self, name)



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import json
import logging
from pathlib import Path
from typing import Callable, Dict, Iterator

from .tag import Tag
from .util import SubDir


LOG = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------
class Branch(SubDir):
    """
    TODO
    """

    ROOT_DIR = "branches"

    @classmethod
    def parent_path(cls, parent: "ArtifactRepository") -> Path:
        return parent.path() / cls.ROOT_DIR

    def url_bit(self) -> str:
        return f"branches/{self.name}"

    def get_tags(self, access_cb: Callable[[str], bool] = None) -> Iterator[Tag]:
        for t in Tag.all(self):
            if access_cb is not None and not access_cb(t.url()):
                continue
            yield t

    def get_tag(self, name: str, access_cb: Callable[[str], bool] = None) -> Tag:
        if name in ("latest", "stable", "oldstable"):
            tags = list(self.get_tags(access_cb))
            tags.sort(key=Tag.creation_date, reverse=True)
            first = True
            for t in tags:
                if name == "latest":
                    return t
                if t.is_stable():
                    if name == "stable" or not first:
                        return t
                    first = False
            raise FileNotFoundError(name)
        return Tag(self, name)

    def _policy_path(self) -> Path:
        return self._path / ".cleanup_policy"

    def set_cleanup_policy(self, max_daily_tags: int, max_released_tags: int):
        policy = {
            "max_daily_tags": max_daily_tags,
            "max_released_tags": max_released_tags,
        }
        self._policy_path().write_text(json.dumps(policy))

    def get_cleanup_policy(self) -> Dict[str, int]:
        try:
            policy = json.loads(self._policy_path().read_text())
        except (OSError, ValueError):
            policy = {}
        for field in "max_daily_tags", "max_released_tags":
            if field not in policy:
                policy[field] = 0
        return policy

    def delete(self, *, force: bool = False):
        if not self.exists():
            raise FileNotFoundError()
        for t in self.get_tags():
            if t.is_locked():
                raise OSError(f"Tag {t.name} is locked")
            if not force and t.is_released():
                raise OSError(f"Tag {t.name} is released, use force")
        for t in self.get_tags():
            t.delete(force=force)
        self.root().rmtree(self._path)



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio

from aiohttp import web
import aiohttp_jinja2

from .util import BaseView


# --------------------------------------------------------------------------------------
class BranchesView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches"
        yield "/branches/"
        yield "/~{user}/branches"
        yield "/~{user}/branches/"

    async def get(self):
        """
        Get the list of branches.
        """
        if not self.request.path.endswith("/"):
            raise web.HTTPFound(self.request.path + "/")
        data = {"branches": []}
        for b in self.repo().get_branches():
            if not self.access_granted(b.url()):
                continue
            br = {
                "name": b.name,
                "daily_tags": 0,
                "released_tags": 0,
            }
            for t in b.get_tags():
                if t.is_released():
                    br["released_tags"] += 1
                else:
                    br["daily_tags"] += 1
            br.update(b.get_cleanup_policy())
            data["branches"].append(br)
        if "html" in self.request.headers.get("Accept", "json"):
            return aiohttp_jinja2.render_template("branches.html", self.request, data)
        return web.json_response(data)


# --------------------------------------------------------------------------------------
class BranchView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}"
        yield "/branches/{branch}/"
        yield "/~{user}/branches/{branch}"
        yield "/~{user}/branches/{branch}/"

    def _get_branch(self):
        branch = self.repo().get_branch(self.request.match_info["branch"])
        if not branch.exists():
            raise web.HTTPNotFound()
        if not self.request.path.endswith("/"):
            raise web.HTTPFound(self.request.path + "/")
        return branch

    async def get(self):
        """
        Get the list of tags for a branch.
        """
        branch = self._get_branch()
        released_only = bool(self.request.query.get("released"))
        data = {
            "branch": {
                "name": self.request.match_info["branch"],
                "cleanup_policy": branch.get_cleanup_policy(),
                "tags": [],
            },
        }
        for t in branch.get_tags():
            if not self.access_granted(t.url()):
                continue
            if released_only and not t.is_released():
                continue
            data["branch"]["tags"].append(
                {
                    "name": t.name,
                    "timestamp": t.timestamp,
                    "released": t.is_released(),
                    "locked": t.is_locked(),
                    "stable": t.is_stable(),
                    "publish_status": t.publish_status(),
                }
            )
        if "html" in self.request.headers.get("Accept", "json"):
            return aiohttp_jinja2.render_template("branch.html", self.request, data)
        return web.json_response(data)

    async def put(self):
        """
        Update the cleanup policy for a branch.
        """
        branch = self._get_branch()
        prev = branch.get_cleanup_policy()
        try:
            policy = (await self.json_body())["branch"]
            max_daily = policy.get("max_daily_tags", prev["max_daily_tags"])
            max_released = policy.get("max_released_tags", prev["max_released_tags"])
            if not (isinstance(max_daily, int) and isinstance(max_released, int)):
                raise TypeError()
            if max_daily < 0 or max_released < 0:
                raise TypeError()
        except (TypeError, KeyError) as e:
            raise web.HTTPBadRequest(reason="invalid parameters") from e
        branch.set_cleanup_policy(max_daily, max_released)
        return web.Response()

    async def delete(self):
        """
        Delete a branch and all its contents.
        """
        loop = asyncio.get_running_loop()
        branch = self._get_branch()
        try:
            force = "force" in self.request.query
            await loop.run_in_executor(None, lambda: branch.delete(force=force))
            self.repo().schedule_cleanup_orphans()
            return web.Response()
        except OSError as e:
            raise web.HTTPBadRequest(reason=str(e)) from e



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

from http import HTTPStatus
import json
import os

from aiohttp import web


# --------------------------------------------------------------------------------------
class AuthenticationRequired(web.HTTPUnauthorized):

    REALM = os.getenv("DLREPO_AUTH_REALM", "dlrepo")
    AUTH_HEADERS = {
        "WWW-Authenticate": f'Basic realm="{REALM}", charset="utf-8"',
    }

    def __init__(self):
        super().__init__(headers=self.AUTH_HEADERS)


# --------------------------------------------------------------------------------------
class ContainerRegistryError(web.HTTPError):
    docker_code = None
    message = None
    headers = None

    def __init__(self, message=None):
        body = json.dumps(
            {
                "errors": [
                    {
                        "code": self.docker_code,
                        "message": message or self.message or self.docker_code,
                    }
                ]
            }
        ).encode("utf-8")
        h = {"Docker-Distribution-Api-Version": "registry/2.0"}
        if self.headers is not None:
            h.update(self.headers)
        super().__init__(body=body, content_type="application/json", headers=h)


class NameUnknown(ContainerRegistryError):
    status_code = HTTPStatus.NOT_FOUND
    docker_code = "NAME_UNKNOWN"
    message = "Repository name not known to registry."


class ManifestUnknown(ContainerRegistryError):
    status_code = HTTPStatus.NOT_FOUND
    docker_code = "MANIFEST_UNKNOWN"
    message = "Unknown manifest in registry."


class ManifestBlobUnknown(ContainerRegistryError):
    status_code = HTTPStatus.NOT_FOUND
    docker_code = "MANIFEST_BLOB_UNKNOWN"
    message = "Unknown blob in registry."


class BlobUploadInvalid(ContainerRegistryError):
    status_code = HTTPStatus.BAD_REQUEST
    docker_code = "BLOB_UPLOAD_INVALID"
    message = "Blob upload invalid."


class BlobUploadUnknown(ContainerRegistryError):
    status_code = HTTPStatus.NOT_FOUND
    docker_code = "BLOB_UPLOAD_UNKNOWN"
    message = "Blob upload unknown to registry."


class BlobUnknown(ContainerRegistryError):
    status_code = HTTPStatus.NOT_FOUND
    docker_code = "BLOB_UNKNOWN"
    message = "Unknown blob in registry."


class Unsupported(ContainerRegistryError):
    status_code = HTTPStatus.BAD_REQUEST
    docker_code = "UNSUPPORTED"
    message = "The operation is unsupported."


class ManifestInvalid(ContainerRegistryError):
    status_code = HTTPStatus.BAD_REQUEST
    docker_code = "MANIFEST_INVALID"
    message = "Invalid manifest."


class Denied(ContainerRegistryError):
    status_code = HTTPStatus.FORBIDDEN
    docker_code = "DENIED"
    message = "Requested access to the resource is denied."


class Unauthorized(ContainerRegistryError):
    status_code = HTTPStatus.UNAUTHORIZED
    docker_code = "UNAUTHORIZED"
    message = "Authentication is required."
    headers = AuthenticationRequired.AUTH_HEADERS



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
import logging.config
import os
import signal
import socket
import sys

from aiohttp import web

from . import views
from .fs import ArtifactRepository
from .views import auth


# --------------------------------------------------------------------------------------
class AccessLogger(web.AbstractAccessLogger):
    def log(self, request: web.BaseRequest, response: web.Response, time: float):
        self.logger.debug(
            "%s@%s - %s %s %s %r",
            request.get("dlrepo_user"),
            request.headers.get("X-Forwarded-For", request.remote),
            response.status,
            request.method,
            request.path_qs,
            request.headers.get("User-Agent"),
        )


# --------------------------------------------------------------------------------------
async def create_semaphore(app: web.Application):
    max_requests = int(os.getenv("DLREPO_PUBLISH_MAX_REQUESTS", "1"))
    app["dlrepo_publish_semaphore"] = asyncio.Semaphore(max_requests)


# --------------------------------------------------------------------------------------
async def init_fs_cleanup(app: web.Application):
    repo = app["dlrepo_artifact_repository"]
    repo.start_cleanup_timer()
    loop = asyncio.get_running_loop()
    loop.add_signal_handler(signal.SIGUSR1, repo.schedule_cleanup_all)


# --------------------------------------------------------------------------------------
def main():
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "console": {
                    "format": "%(levelname)s: %(message)s",
                },
                "syslog": {
                    "format": "dlrepo: %(message)s",
                },
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "console",
                },
                "syslog": {
                    "class": "logging.handlers.SysLogHandler",
                    "address": "/dev/log",
                    "formatter": "syslog",
                },
            },
            "root": {
                "level": os.getenv("DLREPO_LOG_LEVEL", "WARNING"),
                "handlers": [
                    os.getenv(
                        "DLREPO_LOG_OUTPUT",
                        "console" if sys.stdin.isatty() else "syslog",
                    ),
                ],
            },
        }
    )
    app = web.Application(middlewares=[auth.middleware])
    app[auth.AuthBackend.KEY] = auth.AuthBackend()
    app.on_startup.append(app[auth.AuthBackend.KEY].init)

    repo = ArtifactRepository(os.getenv("DLREPO_ROOT_DIR", "/var/lib/dlrepo"))
    app["dlrepo_artifact_repository"] = repo

    app.on_startup.append(create_semaphore)
    app.on_startup.append(init_fs_cleanup)

    views.add_routes(app)

    kwargs = {"access_log_class": AccessLogger}

    # try inherit from systemd socket
    listen_fds = int(os.getenv("LISTEN_FDS", "0"))
    listen_pid = int(os.environ.get("LISTEN_PID", "0"))
    if listen_fds == 1 and listen_pid == os.getpid():
        kwargs["sock"] = socket.fromfd(3, socket.AF_UNIX, socket.SOCK_STREAM)
        kwargs["sock"].setblocking(False)
    else:
        # no systemd provided socket, listen on localhost by default
        kwargs["host"] = os.getenv("DLREPO_LISTEN_ADDRESS", "127.0.0.1")
        kwargs["port"] = int(os.getenv("DLREPO_LISTEN_PORT", "1337"))

    web.run_app(app, **kwargs)


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import logging
from typing import Callable

from aiohttp import web

from .util import BaseView, TarResponse


LOG = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------
class FormatDirView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}/{tag}/{job}/{format}/"
        yield "/~{user}/branches/{branch}/{tag}/{job}/{format}/"
        yield "/products/{product}/{variant}/{product_branch}/{version}/{format}/"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/{version}/{format}/"

    async def head(self):
        fmt = _get_format(self.repo(), self.request.match_info, self.access_granted)
        if fmt.is_dirty():
            raise web.HTTPNotFound()
        if fmt.url() != self.request.path:
            raise web.HTTPFound(fmt.url())
        return web.Response()

    async def get(self):
        """
        Get the list of files of a job for the specified format.
        """
        fmt = _get_format(self.repo(), self.request.match_info, self.access_granted)
        if fmt.url() != self.request.path:
            raise web.HTTPFound(fmt.url())
        if "html" in self.request.headers.get("Accept", "json"):
            return await self.autoindex(fmt, "")
        data = {
            "artifact_format": {
                "name": fmt.name,
                "dirty": fmt.is_dirty(),
                "files": list(fmt.get_digests().keys()),
            },
        }
        return web.json_response(data)

    async def patch(self):
        """
        Remove the dirty flag from a format.
        """
        version = self.request.match_info.get(
            "tag", self.request.match_info.get("version")
        )
        if "product" in self.request.match_info or version in (
            "latest",
            "stable",
            "oldstable",
        ):
            raise web.HTTPMethodNotAllowed("PATCH", ["GET"])
        fmt = _get_format(self.repo(), self.request.match_info)
        try:
            await fmt.post_process()
            fmt.set_dirty(False)
        except OSError as e:
            LOG.error("post process failed: %s", e)
            raise web.HTTPInternalServerError(reason="post process failed") from e
        return web.Response()


# --------------------------------------------------------------------------------------
class FormatArchiveView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}/{tag}/{job}/{format}.tar"
        yield "/~{user}/branches/{branch}/{tag}/{job}/{format}.tar"
        yield "/products/{product}/{variant}/{product_branch}/{version}/{format}.tar"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/{version}/{format}.tar"

    async def head(self):
        fmt = _get_format(self.repo(), self.request.match_info, self.access_granted)
        if fmt.is_dirty():
            raise web.HTTPNotFound()
        url = fmt.url().rstrip("/") + ".tar"
        if url != self.request.path:
            raise web.HTTPFound(url)
        return web.Response()

    async def get(self):
        fmt = _get_format(self.repo(), self.request.match_info, self.access_granted)
        url = fmt.url().rstrip("/") + ".tar"
        if url != self.request.path:
            raise web.HTTPFound(url)

        return TarResponse(
            fmt.get_digests(), fmt.path(), fmt.archive_name(), fmt.timestamp()
        )


# --------------------------------------------------------------------------------------
class FormatDigestsView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}/{tag}/{job}/{format}.sha256"
        yield "/~{user}/branches/{branch}/{tag}/{job}/{format}.sha256"
        yield "/products/{product}/{variant}/{product_branch}/{version}/{format}.sha256"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/{version}/{format}.sha256"

    async def get(self):
        fmt = _get_format(self.repo(), self.request.match_info, self.access_granted)
        url = fmt.url().rstrip("/") + ".sha256"
        if url != self.request.path:
            raise web.HTTPFound(url)
        sha256sums = []
        for artifact, digest in sorted(fmt.get_digests().items()):
            algo, digest = digest.split(":")
            if algo == "sha256":
                sha256sums.append(f"{digest}  {artifact}\n")
        return web.Response(text="".join(sha256sums))


# --------------------------------------------------------------------------------------
class FormatFileView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}/{tag}/{job}/{format}"
        yield "/~{user}/branches/{branch}/{tag}/{job}/{format}"
        yield "/products/{product}/{variant}/{product_branch}/{version}/{format}"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/{version}/{format}"

    async def head(self):
        fmt = _get_format(self.repo(), self.request.match_info, self.access_granted)
        if fmt.is_dirty():
            raise web.HTTPNotFound()
        url = fmt.url()
        files = list(fmt.get_digests().keys())
        if len(files) == 1:
            url += files[0]
        if url != self.request.path:
            raise web.HTTPFound(url)
        return web.Response()

    async def get(self):
        """
        If only one file in $format:
            redirect to /branches/$branch/$tag/$job/$format/$file
        else:
            redirect to /branches/$branch/$tag/$job/$format/
        """
        fmt = _get_format(self.repo(), self.request.match_info, self.access_granted)
        files = list(fmt.get_digests().keys())
        if len(files) == 1 and files[0] != "index.html":
            return web.HTTPFound(fmt.url() + files[0])
        return web.HTTPFound(fmt.url())


# --------------------------------------------------------------------------------------
def _get_format(repo, match_info, access_cb: Callable[[str], bool] = None):
    try:
        if "product" in match_info:
            fmt = (
                repo.get_product(match_info["product"])
                .get_variant(match_info["variant"])
                .get_branch(match_info["product_branch"])
                .get_version(match_info["version"], access_cb)
                .get_format(match_info["format"])
            )
        else:
            fmt = (
                repo.get_branch(match_info["branch"])
                .get_tag(match_info["tag"], access_cb)
                .get_job(match_info["job"])
                .get_format(match_info["format"])
            )
    except FileNotFoundError as e:
        raise web.HTTPNotFound() from e
    if not fmt.exists():
        raise web.HTTPNotFound()
    return fmt
